/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.genclass;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Collection;

import javax.jdo.JDOUserException;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.perseus.cache.api.FixableCacheEntry;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.genclass.api.SpeedoGenClassCoherence;
import org.objectweb.speedo.genclass.api.SpeedoGenClassProxy;
import org.objectweb.speedo.mim.api.SpeedoAccessor;
import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.speedo.mim.lib.BasicSpeedoAccessor;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class represents an accessor of a genclass structure. This accessor is
 * used in case of the generic class is persistent.
 * 
 * @author S.Chassande-Barrioz
 */
public abstract class GenClassAccessor
	extends BasicSpeedoAccessor
	implements PGenClassAccessor, SpeedoGenClassCoherence {

	/**
	 * The SpeedoProxy attached to this state representation.
	 */
	public SpeedoGenClassProxy jdoProxy;

	/**
	 * Indicates if this accessor support the dela mode
	 */
	public boolean supportDelta = true;

	/**
	 * A temporary variable which is used to search an element into the
	 * "elements" list.
	 */
	protected GenClassElement tmpelem = null;

	protected Logger logger;

	/**
	 * The ArrayList used to store the indexed elements of the genclass. Then
	 * this list contains also the deleted elements.
	 */
	public List elements = null;

	public GenClassAccessor(SpeedoGenClassProxy _jdoProxy) {
		elements = new ArrayList();
		this.jdoProxy = _jdoProxy;
		this.proxy = jdoProxy;
		this.tmpelem = (GenClassElement) createPIndexedElem();
	}

	public abstract void loadFieldsFromAccessor(SpeedoAccessor sa);
	
	public abstract PIndexedElem createPIndexedElem(GenClassAccessor gca);

	public abstract void setElements(Object o);

	public Logger getLogger() {
		if (logger == null) {
			logger = ((Loggable) jdoProxy).getLogger();
		}
		return logger;
	}

	public void makePersistent(ProxyManager pm, SpeedoProxy sp) {
        if (!sp.jdoIsActive()) {
            if (sp instanceof SpeedoGenClassProxy) {
				//Genclass contains genclass instances
                SpeedoGenClassProxy sgcp = (SpeedoGenClassProxy) sp;
                if (sgcp.jdoGetPType() == null) {
                    //Assign the PType to the generic class
                    sgcp.jdoSetPType(jdoProxy.jdoGetPType().getNestedPType());
                }
                sgcp.jdoSetLinkedField(jdoProxy.jdoGetGenClassId());
                sgcp.jdoSetPNameHints(jdoProxy.getPName());
            }
			if (Debug.ON) {
				logger.log(BasicLevel.DEBUG, "jdoMakePersistent() pm=" + pm);
			}
			if (pm == null) {
				pm = (ProxyManager) jdoProxy.jdoGetPersistenceManager();
			}
			if (pm == null) {
				throw new JDOUserException(
					"When a persistent object is used (read or write), " +
					"a PersistenceManager is needed");
			}
            pm.speedoMakePersistent(sp, true);
        }
    }

    public void makePersistent(ProxyManager pm, Iterator it) {
        if (it == null) {
            return;
        }
        while(it.hasNext()) {
            Object o = it.next();
            if (o instanceof SpeedoProxy) {
                makePersistent(pm, (SpeedoProxy) o);
            } else {
                // Optimisation: If the first element is not a SpeedoProxy, that
                // means that ALL collection elements are not SpeedoProxy
                // implementation
                return;
            }
        }
    }

	/**
	 * When the gen class is flushed to the data support (parameter = true),
	 * the internal structure is clean: - really remove the deleted element -
	 * mark the genclass as supporting the delta
	 */
	public void setFlushed(boolean val) {
		super.setFlushed(val);
		if (val == true) {
			supportDelta = true;
			Iterator it = elements.iterator();
			while (it.hasNext()) {
				GenClassElement gce = (GenClassElement) it.next();
				if (gce.getElemStatus() == PIndexedElem.ELEM_DELETED) {
					//Remove the deleted element
					it.remove();
				}
			}
		}
	}

	/**
	 * At the end of the working set, the writing mode is initialized
	 * (deltaSupported), and the reference to the persistent are unswilled
	 */
	public void workingSetClosed() {
		if (Debug.ON) {
			((Loggable) jdoProxy).getLogger().log(
				BasicLevel.DEBUG,
				"workingSetClosed()");
		}
		supportDelta = true;
		Iterator it = elements.iterator();
		boolean unswizzle =
			((FixableCacheEntry) getCacheEntry()).getCeFixCount() == 1;
		while (it.hasNext()) {
			GenClassElement gce = (GenClassElement) it.next();
			if (unswizzle) {
				gce.unSwizzle();
			}
			if (gce.getElemStatus() == PIndexedElem.ELEM_DELETED) {
				//Remove the deleted element
				it.remove();
			}
		}
	}

	public void loadFields(ProxyManager pm, long[] fields) {
	}

	public abstract void deletePersistent(ProxyManager pm);

	public abstract void makePersistent(ProxyManager pm);

	public abstract void detachCopy(ProxyManager pm, Map map, SpeedoAccessor fieldsClone,  Collection fgHints);
	
	public abstract void attachCopy(ProxyManager pm, Map map, SpeedoAccessor fieldsClone, boolean makeTransactional);
	
	public abstract void refresh(ProxyManager pm, Map map, Collection fgHints);
	
	public abstract void retrieve(ProxyManager pm, Map map, Collection fgHints);
	
	// IMPLEMENTATION OF THE SpeedoGenClassCoherence INTERFACE //
	//---------------------------------------------------------//

	public abstract boolean speedoAdd(Object elemToAdd, Object hints);

	public abstract boolean speedoRemove(Object elemToRemove, Object hints);

	// IMPLEMENTATION OF THE PGenClassAccessor INTERFACE //
	//---------------------------------------------------//

	public PIndexedElem createPIndexedElem() {
		return createPIndexedElem(this);
	}

	public Object getMemoryInstance() {
		return jdoProxy;
	}

	public void paAdd(PIndexedElem elem, Object conn) throws PExceptionIO {
		// the elem is read from the DS, set it to unmodified.
		 ((GenClassElement) elem).setStatus(PIndexedElem.ELEM_UNMODIFIED);
		elements.add(elem);
	}

	public int paGetNbElem() {
		return elements.size();
	}

	public Iterator paIterator() {
		return elements.iterator();
	}

	public boolean paDeltaSupported() {
		return supportDelta;
	}

	public void paSetNbElem(int nbelem) {
		elements = (nbelem < 0 ? new ArrayList() : new ArrayList(nbelem));
	}

}
