/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */


package org.objectweb.speedo.genclass;

import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.speedo.mim.api.SpeedoAccessor;
import org.objectweb.speedo.genclass.api.SpeedoGenClassProxy;

import javax.jdo.PersistenceManager;
import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * This class is an implementation of the iterator interface. It iterates over
 * an inner iterator which the elements are GenClassElement. The main roles of
 * this iterator implementation are
 * - to skip the GenClassElement marked as deleted
 * - to resolve PName references if the element are persistent object
 *
 * @author Sebastien Chassande-Barrioz
 */
public class PIndexedElemIterator implements Iterator {

    /**
     * Is the inner iterator over PIndexexElem instances
     */
    protected Iterator iter;

    /**
     * is the next element which has not been deleted
     */
    protected GenClassElement next;

    /**
     * indicates if the next element has been computed
     */
    protected boolean nextComputed;

    /**
     * is the object to synchronize if an element must be removed
     */
    protected Object synchro = null;

    /**
     * is the persistence manager used to resolve the PName into reference.
     * If this field has a null value, that means the elements of the gen class
     * are not references but primitives.
     */
    protected PersistenceManager pm = null;

    protected Logger logger = null;


    public PIndexedElemIterator(Collection _elements, Object _synchro,
            PersistenceManager pm, Logger l) {
        iter = _elements.iterator();
        synchro = _synchro;
        this.pm = pm;
        logger = l;
    }


    // IMPLEMENTATION OF THE Iterator INTERFACE //
    //------------------------------------------//

    public boolean hasNext() {
        if (!nextComputed) {
            while (iter.hasNext()) {
                next = (GenClassElement) iter.next();
                if (next.getElemStatus() != PIndexedElem.ELEM_DELETED) {
                    nextComputed = true;
                    return true;
                }
            }
            next = null;
            nextComputed = true;
            return false;
        } else {
            return next != null;
        }
    }

    public Object next() {
        if (!nextComputed) {
            hasNext();
        }
        if (next == null) {
            throw new NoSuchElementException();
        } else {
            nextComputed = false;
            return (pm == null ? next.getElement() : next.getElement(pm));
        }
    }

    public void remove() {
        synchronized (synchro) {
            if (next != null) {
                SpeedoAccessor sa = next.getSpeedoAccessor();
                sa.getSpeedoProxy().getSpeedoHome().writeIntention(sa.getSpeedoProxy(), null);
                next.setStatus(PIndexedElem.ELEM_DELETED);
            }
        }
    }
}
