/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.genclass.merger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.taskdefs.MatchingTask;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeAdapter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.genclass.GenClass;
import org.objectweb.speedo.generation.enhancer.ClassRenamer;
import org.objectweb.speedo.tools.StringReplace;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class GenClassMerger
extends MatchingTask {

	public final static String LOGGER_NAME =
		SpeedoProperties.LOGGER_NAME + ".genclassmerger";

	public final static String GEN_CLASS_NAME
	= GenClass.class.getName().replace('.', '/');


	public final static String[][] CONSTRUCTOR_PARAM = {
			{"Lorg/objectweb/jorm/type/api/PType;", "ptype"},
			{"Ljava/lang/String;", "linkedField"},
			{"Ljava/lang/Object;", "pnameHints"},
			{"Ljava/lang/String;", "mapperName"}
	};

	public final static String CONSTRUCTOR_DESC =";"
		+ CONSTRUCTOR_PARAM[0][0]
							   + CONSTRUCTOR_PARAM[1][0]
													  + CONSTRUCTOR_PARAM[2][0]
																			 + CONSTRUCTOR_PARAM[3][0]
																									+ ")V";

	private File src = null;

	private Logger logger = null;

	public GenClassMerger() {
	}

	public GenClassMerger(Logger logger) {
		this.logger = logger;
	}

	public void _setLogger(Logger logger) {
		this.logger = logger;
	}

	public File getSrc() {
		return src;
	}

	public void setSrc(File src) {
		this.src = src;
	}

	public void execute() throws BuildException {
		if (logger == null) {
			logger = Monolog.initialize().getLogger(LOGGER_NAME);
		}
		String[] pdFiles = super.getDirectoryScanner(src).getIncludedFiles();
		for(int i=0; i<pdFiles.length; i++) {
			try {
				mergeGenClass(pdFiles[i]);
			} catch (SpeedoException e) {
				String msg = "Error while merging the file " + pdFiles[i]
																	   + " merged" ;
				Exception ie = ExceptionHelper.getNested(e);
				logger.log(BasicLevel.ERROR, msg, ie);
				throw new BuildException(msg, ie);
			}
		}
	}

	public void mergeGenClass(String gcn) throws SpeedoException {
		IsAbstractVisitor i = new IsAbstractVisitor();
		ClassReader cr = loadJavaClass(gcn, false);
		cr.accept(i, false);
		if (!i.isAbstract) {
			return;
		}
		logger.log(BasicLevel.INFO, "Enhance the generic class " + gcn);

		ClassWriter cw = new ClassWriter(false);

		//Write the original classes
		ArrayList methods = new ArrayList();
		ClassMerger m = new ClassMerger(cw, methods, null);
		cr.accept(m, false);

		//Write the genclass classes
		Set old = new TreeSet();
		String t = gcn.substring(0, gcn.length()-6);
		
		t = StringReplace.replaceChar(File.separatorChar, '/', t);
		logger.log(BasicLevel.DEBUG, "t=" + t);
		
		old.add(t);
		m = new ClassMerger(cw,methods, m.getSuperName());
		ClassRenamer renamer = new ClassRenamer(m, old, t, new LoggerImpl());
		loadJavaClass(GEN_CLASS_NAME  + ".class", false).accept(renamer, false);
		writeJavaClass(gcn, cw);
		try {
			t = StringReplace.replaceChar('/','.', t);
			logger.log(BasicLevel.DEBUG, "className=" + t);
			Class.forName(t);
			
		} catch (Exception e) {
			String msg = "Bad merge of the class " + gcn;
			logger.log(BasicLevel.ERROR, msg, e);
			throw new SpeedoException(msg, e);
		}
	}

	/**
	 * This visitor checks if the class is abstract.
	 */
	private class IsAbstractVisitor implements ClassVisitor {
		public boolean isAbstract;

		public void visit(final int access,
				final String name,
				final String superName,
				final String[] interfaces,
				final String sourceFile) {
			isAbstract = (access & Constants.ACC_ABSTRACT) != 0;
		}

		public void visitInnerClass(String s, String s1, String s2, int i) {
		}

		public void visitField(int i, String s, String s1, Object o, Attribute attrs) {
		}

		public CodeVisitor visitMethod(int i, String s, String s1, String[] strings, Attribute attrs) {
			return null;
		}

		public void visitAttribute(Attribute attribute) {
		}

		public void visitEnd() {
		}
	}


	/**
	 * Loads a specified class.
	 * @param filename is the file name of the .class to load. the file name is
	 * a relative patht to the 'src' directory.
	 * @param remove indicates if the .class must be removed
	 * @return the JavaClass loaded
	 * @exception SpeedoException if the file cannot be loaded
	 */
	protected ClassReader loadJavaClass(String filename,
			boolean remove) throws SpeedoException {
		try {
			File f = new File(src, filename);
			FileInputStream fis = new FileInputStream(f);
			ClassReader jclass = new ClassReader(fis);
			fis.close();
			if (remove) {
				f.delete();
			}
			return jclass;
		} catch (IOException e) {
			throw new SpeedoException("Error during loading " + filename, e);
		}
	}

	/**
	 * Saves the new bytecode of the specified Java class under a specified base
	 * directory.
	 *
	 * @param jclass the Java class that has to be saved
	 * @exception SpeedoException if the file cannot be written
	 */
	protected void writeJavaClass(final String filename,
			final ClassWriter jclass)
	throws SpeedoException {
		try {
			File outputFile = new File(src, filename);
			outputFile.createNewFile();
			FileOutputStream fos = new FileOutputStream(outputFile);
			fos.write(jclass.toByteArray());
			fos.close();
		} catch (IOException e) {
			throw new SpeedoException("Cannot write " + filename, e);
		}
	}

	/**
	 * 1/ Writes the non abstract methods and the fields of a class.
	 * 2/ Add in the special constructor the field assignement.
	 */
	private class ClassMerger extends ClassAdapter {

		List methods;
		String superName;
		String classname;

		public ClassMerger(ClassVisitor classVisitor,
				List methods,
				String superName) {
			super(classVisitor);
			this.methods = methods;
			this.superName = superName;
		}


		public String getSuperName() {
			return superName;
		}

		/**
		 * Makes the classes no more abstract
		 */
		public void visit(final int access,
				final String name,
				final String superName,
				final String[] interfaces,
				final String sourceFile) {
			this.classname = name;
			this.superName = (this.superName == null ? superName : this.superName);
			super.cv.visit(access ^ Constants.ACC_ABSTRACT,
					name, this.superName, interfaces, sourceFile);
		}

		public CodeVisitor visitMethod(final int access,
				final String name,
				final String desc,
				final String[] exceptions,
				final Attribute attrs) {
			String method = name + " " +  desc;
			if ((access & Constants.ACC_ABSTRACT) == 0
					&& !methods.contains(method)) {
				methods.add(method);

				CodeVisitor cov = super.cv.visitMethod(access, name, desc, exceptions, attrs);
				if (name.equals("<init>") && desc.endsWith(CONSTRUCTOR_DESC)) {
					cov = new ConstructorModifier(cov, classname);
				}
				return cov;
			} else {
				return null;
			}
		}

		private class ConstructorModifier extends CodeAdapter {
			String className;
			public ConstructorModifier(CodeVisitor codeVisitor, String classname) {
				super(codeVisitor);
				this.className = classname;
			}

			public void visitInsn(int i) {
				if (i == Constants.RETURN) {
					for(int ps = 0; ps<CONSTRUCTOR_PARAM.length; ps++) {
						//ALOAD 0
						cv.visitVarInsn(Constants.ALOAD, 0);
						//ALOAD X
						cv.visitVarInsn(Constants.ALOAD, ps + 2);
						//PUTFIELD Y
						cv.visitFieldInsn(Constants.PUTFIELD,
								className, CONSTRUCTOR_PARAM[ps][1],
								CONSTRUCTOR_PARAM[ps][0]);
					}
				}
				super.visitInsn(i);
			}
		}
	}
}
