/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.Attribute;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoModifier;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.Collection;
import java.util.Iterator;

/**
 * Adds getter and setter methods to a class.
 *
 * Adapted from generateAccessors, genGetter and genSetter in EnhancerTool.
 */
public class ClassAccessorAdder extends LoggedClassAdapter {

    /**
     * The Speedo meta information for the visited class.
     */
    private final SpeedoClass jdoClass;

    /**
     * A collection of SpeedoXMLDescriptor describing known persistent classes
     */
    private final Collection xmlDescriptors;

    /**
     * Creates a new {@link ClassAccessorAdder}.
     *
     * @param _cv the class visitor to be used to generate the modified class
     * @param _jdoClass the Speedo meta information for the visited class.
     */
    public ClassAccessorAdder(final ClassVisitor _cv,
                              final SpeedoClass _jdoClass,
                              final Collection _xmlDescriptors,
                              Logger _logger) {
        super(_cv, _logger);
        this.jdoClass = _jdoClass;
        this.xmlDescriptors = _xmlDescriptors;
    }

    // IMPLEMENTATION OF THE ClassVisitor INTERFACE //
    // ---------------------------------------------//

    public void visit(final int access,
                      final String name,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        cv.visit(access, name, superName, interfaces, sourceFile);

        // Creates usable accessors for each jdo field
        Iterator i = jdoClass.jdoField.values().iterator();
        while (i.hasNext()) {
            SpeedoField jdoField = (SpeedoField) i.next();
            if (jdoField.name == null) {
                continue;
            }
            int acc = jdoField.access;
            if (jdoField.persistenceModifier >= SpeedoModifier.transactional) {
                // fields that are persistent or transactional
                if ((acc & Constants.ACC_STATIC) != 0) {
                    throw new RuntimeException(
                            "A static field cannot be defined as persistent");
                }
                genSetter(name, jdoField);
                if (jdoField.persistenceModifier == SpeedoModifier.persistent) {
                    genGetter(name, jdoField);
                }
            } else if ((acc &
                    (Constants.ACC_TRANSIENT | Constants.ACC_FINAL)) != 0) {
                // transient and/or final field
                genSetter(name, jdoField);
                genGetter(name, jdoField);
            } else if (jdoField.persistenceModifier == SpeedoModifier.missing) {
                if (Util.isPersistentType(jdoField.desc, xmlDescriptors)) {
                    // fields whose persistent-modifier isn't defined
                    genSetter(name, jdoField);
                    genGetter(name, jdoField);
                }
            }
        }
    }

    public CodeVisitor visitMethod(final int access,
                                   final String name,
                                   final String desc,
                                   final String[] exceptions,
                                   final Attribute attrs) {
        // optimization: do not create an empty code adapter for visited
        // methods, as is done by default in ClassAdapter
        return cv.visitMethod(access, name, desc, exceptions, attrs);
    }

    // OTHER METHODS //
    // --------------//

    /**
     * Generates a getter method for the given field.
     * public float jdoGetF1() {
     *   TotoField sa = (TotoField) this.jdoReadIntention(new long[]{O, ..., 0, x})
     *   return sa.f1
     * }
     * @param owner the internal name of the visited class
     * @param jdoField the field for which the getter method must be generated
     */
    private void genGetter(final String owner, final SpeedoField jdoField) {
        int flags = jdoField.access;
        String gettername =
                NamingRules.getterName(jdoField.jdoClass, jdoField.name);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the getter method '"
                    + gettername + "' on the class '" + jdoField.jdoClass.name + "'");
        }
        String desc = jdoField.desc;
        //generate method signature
        CodeVisitor _cv = this.cv.visitMethod(flags, gettername, "()" + desc, null, null);

        //load this for the readIntention call
        _cv.visitVarInsn(Constants.ALOAD, 0);
        
        // define the long[] value such as new long[]{O, ..., 0, x}
        int idx = (jdoField.number / 64);
        int arraysize = idx + 1;
        Util.visitIntConstant(_cv, arraysize);
        _cv.visitIntInsn(NEWARRAY, 11);
        for(int i=0; i<idx; i++) {
            _cv.visitInsn(DUP);
            Util.visitIntConstant(_cv, i);
            Util.visitLongConstant(_cv, 0);
            _cv.visitInsn(LASTORE);
        }
        _cv.visitInsn(DUP);
        Util.visitIntConstant(_cv, idx);
        Util.visitLongConstant(_cv, 1L << (jdoField.number % 64));
        _cv.visitInsn(LASTORE);

        //call the jdoReadIntention method
        _cv.visitMethodInsn(
                Constants.INVOKEINTERFACE,
                "org/objectweb/speedo/mim/api/SpeedoProxy",
                "jdoReadIntention",
                "(Lorg/objectweb/speedo/mim/api/SpeedoAccessor;[J)V");
        // generate code to return field from fields
        String fieldsOwner = NamingRules.fieldsName(owner);
        _cv.visitVarInsn(Constants.ALOAD, 1);
        _cv.visitTypeInsn(Constants.CHECKCAST, fieldsOwner);
        _cv.visitFieldInsn(Constants.GETFIELD, fieldsOwner, jdoField.name, desc);
        Type returnType = Type.getType(desc);
        _cv.visitInsn(returnType.getOpcode(Constants.IRETURN));
        _cv.visitMaxs(0, 0);
    }

    /**
     * Generates a setter method for the given field.
     *
     * @param owner the internal name of the visited class
     * @param jdoField the field for which the setter method must be generated
     */
    private void genSetter(final String owner, final SpeedoField jdoField) {
        int flags = jdoField.access;
        String name =
                NamingRules.setterName(jdoField.jdoClass, jdoField.name);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the setter method '"
                    + name + "' on the class '" + jdoField.jdoClass.name + "'");
        }
        String desc = jdoField.desc;
        CodeVisitor _cv =
                this.cv.visitMethod(flags, name, "(" + desc + ")V", null, null);
        Type type = Type.getType(desc);
        int params = type.getSize() + 1;
        // generate code to call jdoWriteIntention
        _cv.visitVarInsn(Constants.ALOAD, 0);

        // define the long[] value such as new long[]{O, ..., 0, x}
        int idx = (jdoField.number / 64);
        int arraysize = idx + 1;
        Util.visitIntConstant(_cv, arraysize);
        _cv.visitIntInsn(NEWARRAY, T_LONG);
        for(int i=0; i<idx; i++) {
            _cv.visitInsn(DUP);
            Util.visitIntConstant(_cv, i);
            Util.visitLongConstant(_cv, 0);
            _cv.visitInsn(LASTORE);
        }
        _cv.visitInsn(DUP);
        Util.visitIntConstant(_cv, idx);
        Util.visitLongConstant(_cv, 1L << (jdoField.number % 64));
        _cv.visitInsn(LASTORE);

        _cv.visitMethodInsn(
                Constants.INVOKEINTERFACE,
                "org/objectweb/speedo/mim/api/SpeedoProxy",
                "jdoWriteIntention",
                "[J)V");
        // generate code to set field of state
        String fieldsOwner = NamingRules.fieldsName(owner);
        _cv.visitVarInsn(Constants.ALOAD, params);
        _cv.visitTypeInsn(Constants.CHECKCAST, fieldsOwner);
        _cv.visitVarInsn(type.getOpcode(Constants.ILOAD), 1);
        _cv.visitFieldInsn(Constants.PUTFIELD, fieldsOwner, jdoField.name, desc);
        _cv.visitInsn(Constants.RETURN);
        _cv.visitMaxs(0, 0);
    }
}
