/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.generation.lib.NamingRules;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This class adds/removes dummy accessors in order to permit the compilation.
 * The jdoGetXXX and jdoSetXXX method are added before the compilation and they
 *  are removed after. Indeed these accessors are really generated in the Proxy
 * class, and there merging with the persistent class later.
 *
 * @author S.Chassande-Barrioz
 */
public class DummyAccessorModifier extends LoggedClassAdapter {

	/**
	 * List of method names to removed
	 */
	private Set methods;

	public DummyAccessorModifier(ClassVisitor classVisitor,
								 Logger logger,
								 SpeedoClass sc,
								 boolean beforeCompilation) {
		super(classVisitor, logger);
		if (!beforeCompilation) {
			methods = new HashSet();
		}
		Iterator it = sc.jdoField.values().iterator();
		while (it.hasNext()) {
			SpeedoField sf = (SpeedoField) it.next();
			//if (sf.relationType != SpeedoField.NO_RELATION) {
				String getter = NamingRules.getterName(sc, sf.name);
				String setter = NamingRules.setterName(sc, sf.name);
				String coherencesetter = NamingRules.coherentSetterName(sc, sf.name);
				if (debug) {
					logger.log(BasicLevel.DEBUG,
							"DummyAccessor modifier of the class "
							+ sc.getFQName() + ": getter=" + getter);
				}
				if (beforeCompilation) {
					// add dummy getter
					CodeVisitor _cv = this.cv.visitMethod(ACC_PUBLIC, getter,
							"()" + sf.desc, null, null);
					_cv.visitInsn(ACONST_NULL);
					_cv.visitInsn(ARETURN);
					_cv.visitMaxs(1, 1);
					// add dummy setter
					_cv = this.cv.visitMethod(ACC_PUBLIC, setter,
							"(" + sf.desc + ")V", null, null);
					_cv.visitInsn(RETURN);
					_cv.visitMaxs(2, 2);
					// add dummy coherencesetter
					_cv = this.cv.visitMethod(ACC_PUBLIC, coherencesetter,
							"(" + sf.desc + ")V", null, null);
					_cv.visitInsn(RETURN);
					_cv.visitMaxs(2, 2);
				} else {
					methods.add(getter);
					methods.add(setter);
					methods.add(coherencesetter);
				}
			//}
		}
		if (methods != null && methods.isEmpty()) {
			methods = null;
		}
	}

	public CodeVisitor visitMethod(int modifier,
								   String methodName,
								   String returnType,
								   String[] paramTypes,
								   Attribute attrs) {
		if (methods != null && methods.contains(methodName)) {
			if (debug) {
				logger.log(BasicLevel.DEBUG, "DummyAccessor.visitMethod("
						+ methodName + ") return null");
			}
			// remove dummy getter or setter
			return null;
		} else {
			CodeVisitor _cv = super.visitMethod(
					modifier, methodName, returnType, paramTypes, attrs);
			if (debug) {
				logger.log(BasicLevel.DEBUG, "DummyAccessor.visitMethod("
						+ methodName + ") return cv=" + _cv);
			}
			return _cv;
		}
	}
}
