/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.generation.lib.NamingRules;

/**
 * Modify the inheritance in order to permit the compilation.
 * Before the compilation the parent of the children classes (classes
 * inheriting from another persistent class) are setting to the Parent proxy
 * class. After the compilation the inheritance link is coming back to the
 * real class.<br/>
 * ex: Kangaroo extends Animal <br/>
 *   Action before the compilation: Kangaroo extends AnimalProxy <br/>
 *   Action after the compilation: Kangaroo extends Animal <br/>
 *
 * @author S.Chassande-Barrioz
 */
public class InheritanceModifier extends LoggedClassAdapter {

	/**
	 * The name of the new super class claculated with the meta information.
	 * A null value means that the inheritance does not change.
	 */
	private String newSuperClassName;

	public InheritanceModifier(ClassVisitor classVisitor,
							   Logger logger,
							   SpeedoClass sc,
							   boolean beforeCompilation) {
		super(classVisitor, logger);
		if (sc.superClassName != null
				&& sc.superClassName.length() > 0) {
			newSuperClassName = (beforeCompilation
					? NamingRules.proxyName(sc.superClassName)
					: sc.superClassName);
			newSuperClassName = newSuperClassName.replace('.','/');
		} else {
			newSuperClassName = null;
		}
		if (debug) {
			logger.log(BasicLevel.DEBUG, "Inheritance modifier of the class "
				+ sc.getFQName() + ": newSuperClassName=" + newSuperClassName);
		}
	}

	public void visit(final int access,
					  final String name,
					  final String superName,
					  final String[] interfaces,
					  final String sourceFile) {
		String inherit = (newSuperClassName == null
				? superName : newSuperClassName);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "InheritanceModifier.visit("
					+ name + ", " + superName + "): " + inherit);
		}
		super.visit(access, name, inherit, interfaces, sourceFile);
	}
}
