/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.Attribute;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;
import org.objectweb.speedo.tools.StringReplace;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Add the registration to the JDOImplHelper in a static code area
 *
 * The added code use the class$(String)Class method and the field
 * class$classname which can exist or not. Therefore they are added if they do
 * not already exist.
 *
 * @author S.Chassande-Barrioz
 */
public class JDOImplRegistrationAdder extends LoggedClassAdapter {

	/**
	 * Class name of the current visited class (with slash separator)
	 */
	String className;

	/**
	 * indicates if the field already exists in the current class
	 */
	boolean clinitfound = false;

	/**
	 * indicates if a static area class$classname already exists in the
	 * current class.
	 */
	boolean classFieldfound = false;

	/**
	 * indicates if the class$ method already exists in the current class
	 */
	boolean classMethodfound = false;

	/**
	 * is the name of the static field referencing the java.lang.Class of
	 * the current class.
	 */
	private String fieldName;

	public JDOImplRegistrationAdder(ClassVisitor classVisitor) {
		super(classVisitor);
	}

	public JDOImplRegistrationAdder(ClassVisitor classVisitor, Logger logger) {
		super(classVisitor, logger);
	}

	public void visit(final int access,
					  final String name,
					  final String superName,
					  final String[] interfaces,
					  final String sourceFile) {
		className = name;
		fieldName = "class$" + className.replace('/', '$');
		super.visit(access, name, superName, interfaces, sourceFile);
		logger.log(BasicLevel.DEBUG, "Class name: " + className);
	}

	public void visitField(int i, String s, String s1, Object o, Attribute attribute) {
		super.visitField(i, s, s1, o, attribute);
		classFieldfound |= s.equals(fieldName);
	}

	public CodeVisitor visitMethod(final int access,
								   final String name,
								   final String desc,
								   final String[] exceptions,
								   final Attribute attrs) {
		CodeVisitor c = cv.visitMethod(access, name, desc, exceptions, attrs);
		if (name.equals("<clinit>")) {
			clinitfound = true;
			addClassRegistrationCode(c);
		} else if (name.equals("class$")) {
			classMethodfound = true;
		}
		return c;
	}

	public void visitEnd() {
		if (!clinitfound) {
			logger.log(BasicLevel.DEBUG, "Adding static area in " + className);
			CodeVisitor c = cv.visitMethod(Constants.ACC_STATIC,
				"<clinit>", "()V", null, null);
			addClassRegistrationCode(c);
			c.visitInsn(RETURN);
			c.visitMaxs(0, 0);
		}

		if (!classMethodfound) {
			logger.log(BasicLevel.DEBUG, "Adding static method class$(String)Class in " + className);
			CodeVisitor c = cv.visitMethod(ACC_STATIC + ACC_SYNTHETIC, "class$",
				"(Ljava/lang/String;)Ljava/lang/Class;", null, null);
			Label l0 = new Label();
			c.visitLabel(l0);
			c.visitVarInsn(ALOAD, 0);
			c.visitMethodInsn(INVOKESTATIC, "java/lang/Class", "forName",
				"(Ljava/lang/String;)Ljava/lang/Class;");
			Label l1 = new Label();
			c.visitLabel(l1);
			c.visitInsn(ARETURN);
			Label l2 = new Label();
			c.visitLabel(l2);
			c.visitVarInsn(ASTORE, 1);
			c.visitTypeInsn(NEW, "java/lang/NoClassDefFoundError");
			c.visitInsn(DUP);
			c.visitVarInsn(ALOAD, 1);
			c.visitMethodInsn(INVOKEVIRTUAL, "java/lang/ClassNotFoundException",
				"getMessage", "()Ljava/lang/String;");
			c.visitMethodInsn(INVOKESPECIAL, "java/lang/NoClassDefFoundError",
				"<init>", "(Ljava/lang/String;)V");
			c.visitInsn(ATHROW);
			c.visitTryCatchBlock(l0, l1, l2, "java/lang/ClassNotFoundException");
			c.visitMaxs(3, 2);
		}
		if (!classFieldfound) {
			logger.log(BasicLevel.DEBUG, "Adding static field class$ in " + className);
			cv.visitField(ACC_STATIC + ACC_SYNTHETIC, fieldName,
				"Ljava/lang/Class;", null, null);
		}

		super.visitEnd();
	}

	/**
	 * Add registration code of the class into the JDOImplHelper
	 * @param c
	 */
	private void addClassRegistrationCode(CodeVisitor c) {
		c.visitFieldInsn(Constants.GETSTATIC,
			className,
			fieldName,
			"Ljava/lang/Class;");

		Label l18 = new Label();
		c.visitJumpInsn(Constants.IFNONNULL, l18);
		c.visitLdcInsn(StringReplace.replaceChar('/', '.', className));
		c.visitMethodInsn(Constants.INVOKESTATIC, className, "class$", "(Ljava/lang/String;)Ljava/lang/Class;");
		c.visitInsn(Constants.DUP);
		c.visitFieldInsn(Constants.PUTSTATIC,
			className,
			fieldName,
			"Ljava/lang/Class;");
		Label l21 = new Label();
		c.visitJumpInsn(Constants.GOTO, l21);
		c.visitLabel(l18);
		c.visitFieldInsn(Constants.GETSTATIC,
			className,
			fieldName,
			"Ljava/lang/Class;");
		c.visitLabel(l21);
		c.visitInsn(Constants.ACONST_NULL);
		c.visitInsn(Constants.ACONST_NULL);
		c.visitInsn(Constants.ACONST_NULL);
		c.visitInsn(Constants.ACONST_NULL);
		c.visitInsn(Constants.ACONST_NULL);
		c.visitMethodInsn(Constants.INVOKESTATIC,
			"javax/jdo/spi/JDOImplHelper",
			"registerClass",
			"(Ljava/lang/Class;[Ljava/lang/String;[Ljava/lang/Class;[BLjava/lang/Class;Ljavax/jdo/spi/PersistenceCapable;)V");
	}
}
