/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: speedo@objectweb.org
 *
 */

package org.objectweb.speedo.jca;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.speedo.api.Debug;

import java.util.Collection;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionMetaData;
import javax.resource.cci.Interaction;
import javax.resource.cci.LocalTransaction;
import javax.resource.cci.ResultSetInfo;
import javax.jdo.FetchPlan;
import javax.jdo.JDOConnection;
import javax.jdo.LifecycleListener;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import javax.jdo.Sequence;
import javax.jdo.Transaction;
import javax.jdo.Query;
import javax.jdo.Extent;
import javax.jdo.PersistenceManagerFactory;

/**
 * @author	P. Dechamboux
 */
public class JdoConnection
	implements Connection, ConnectionMetaData, PersistenceManager {
    
    private final static String NO_ACTIVE_CONNECTION = 
        "The JDO connection should be active!!";
    
	/**
	 * The logger into which traces about JdoConnection are produced.
	 */
	private Logger logger;
	/**
	 * The name of this storage sub-system.
	 */
	private final static String EISPRODUCTNAME = "JDO Connector";
	/**
	 * The ConnectionFactory that have requested the allocation of this
	 * Connection.
	 */
	private JdoConnectionFactory connectionFactory;
	/**
	 * The ManagedConnection associated with this Connection.
	 */
	private JdoManagedConnection managedConnection = null;
    /**
     * The LocalTransaction allocated by this connection if any.
     */
    private LocalTransaction localTransaction = null;

	/**
	 * Constructs a JdoConnection.
	 * @param el	The logger into which to produce Connection-related traces.
	 * @param fcf	The ConnectionFactory that has requested the Connection
	 * 				creation.
	 */
    JdoConnection(Logger el, JdoConnectionFactory fcf) {
		logger = el;
		connectionFactory = fcf;
	}

	/**
	 * Initializes a Connection to be linked with a ManagedConnection.
	 * @param mc	The ManagedConnection to be linked to.
	 */
	void initialize(JdoManagedConnection mc) {
		managedConnection = mc;
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)Connection INTERFACE

	/**
	 * No support for Interaction.
	 */
	public Interaction createInteraction() throws ResourceException {
		throw new ResourceException("JDO Connector: no support for Interaction.");
	}

	public LocalTransaction getLocalTransaction() throws ResourceException {
		if (managedConnection == null) {
			throw new ResourceException("JDO Connector: cannot get a LocalTransaction with no associated ManagedConnection.");
        }
        localTransaction = managedConnection;
		return managedConnection;
	}

	/**
	 * The JdoConnection manages the metadata on its own.
	 */
	public ConnectionMetaData getMetaData() throws ResourceException {
		return this;
	}

	/**
	 * No support for ResultSet.
	 */
	public ResultSetInfo getResultSetInfo() throws ResourceException {
		throw new ResourceException("JDO Connector: no support for ResultSet.");
	}

	/**
	 * Closes this Connection. Dissociates from the ManagedConnection to which
	 * it is linked.
	 */
	public void close() {
        if (Debug.ON)
            logger.log(BasicLevel.DEBUG, "Closes Connection: " + this);
		if (managedConnection == null) {
			// The connection is already closed
			return;
		}
        if (localTransaction != null) {
            if (!managedConnection.localTransactionTerminated()) {
                throw new JDOException("JDO Connector: cannot close connection while a LocalTransaction is still active.");
            }
        }
        try {
            managedConnection.dissociateConnection(this);
        } catch (ResourceException e) {
            throw new JDOException("JDO Connector: problem while closing a connection.", e);
        } finally {
            localTransaction = null;
            managedConnection = null;
        }
	}

	/**
	 * Sets the connection to the relevant auto-commit mode.
	 */
	public void setAutoCommit(boolean b) throws ResourceException {
        throw new ResourceException("No sense for a JDO connector.");
	}

	/**
	 * Gets the connection auto-commit mode.
	 */
	public boolean getAutoCommit() throws ResourceException {
        throw new ResourceException("No sense for a JDO connector.");
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)ConnectionMetaData INTERFACE

	/**
	 * Returns the name of this storage sub-system.
	 * @return	The storage sub-system name.
	 */
	public String getEISProductName() throws ResourceException {
		return EISPRODUCTNAME;
	}

	/**
	 * This is the same version number as the adapter.
	 * @return	The storage sub-system version.
	 */
	public String getEISProductVersion() throws ResourceException {
		return connectionFactory.getAdapterVersion();
	}

	/**
	 * No support for user name.
	 * @return	The empty string.
	 */
	public String getUserName() throws ResourceException {
		return "";
	}

	// IMPLEMENTATION OF METHODS FROM THE PersistenceManager INTERFACE
	// Delegates execution to the associated ManagedConnection.

    public boolean isClosed() {
        return managedConnection == null;
    }

    public Transaction currentTransaction() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().currentTransaction();
    }

    public void evict(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().evict(o);
    }

    public void evictAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().evictAll(objects);
    }

    public void evictAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().evictAll(collection);
    }

    public void evictAll() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().evictAll();
    }

    public void refresh(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().refresh(o);
    }

    public void refreshAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().refreshAll(objects);
    }

    public void refreshAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().refreshAll(collection);
    }

    public void refreshAll() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().refreshAll();
    }

    public Query newQuery() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery();
    }

    public Query newQuery(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(o);
    }

    public Query newQuery(String s, Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(s, o);
    }

    public Query newQuery(Class aClass) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(aClass);
    }

    public Query newQuery(Extent extent) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(extent);
    }

    public Query newQuery(Class aClass, Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(aClass, collection);
    }

    public Query newQuery(Class aClass, String s) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(aClass, s);
    }

    public Query newQuery(Class aClass, Collection collection, String s) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(aClass, collection, s);
    }

    public Query newQuery(Extent extent, String s) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newQuery(extent, s);
    }

    public Extent getExtent(Class aClass, boolean b) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getExtent(aClass, b);
    }

    public Object getObjectById(Object o, boolean b) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectById(o, b);
    }

    public Collection getObjectsById(Collection arg0, boolean arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectsById(arg0, arg1);
    }
    public Collection getObjectsById(Collection arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectsById(arg0);
    }
    public Object[] getObjectsById(Object[] arg0, boolean arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectsById(arg0, arg1);
    }
    public Object[] getObjectsById(Object[] arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectsById(arg0);
    }
    public Sequence getSequence(String arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getSequence(arg0);
    }
    public Object getUserObject(Object arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getUserObject(arg0);
    }
    public Object newInstance(Class arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newInstance(arg0);
    }
    public Object putUserObject(Object arg0, Object arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().putUserObject(arg0, arg1);
    }
    public Object removeUserObject(Object arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().removeUserObject(arg0);
    }
    public Object getObjectId(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectId(o);
    }

    public Object getTransactionalObjectId(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getTransactionalObjectId(o);
    }

    public Object newObjectIdInstance(Class aClass, String s) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newObjectIdInstance(aClass, s);
    }

    public void makePersistent(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makePersistent(o);
    }

    public void makePersistentAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makePersistentAll(objects);
    }

    public void makePersistentAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makePersistentAll(collection);
    }

    public void deletePersistent(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().deletePersistent(o);
    }

    public void deletePersistentAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().deletePersistentAll(objects);
    }

    public void deletePersistentAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().deletePersistentAll(collection);
    }

    public void makeTransient(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeTransient(o);
    }

    public void makeTransientAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeTransientAll(objects);
    }

    public void makeTransientAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeTransientAll(collection);
    }

    public void makeTransactional(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeTransactional(o);
    }

    public void makeTransactionalAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeTransactionalAll(objects);
    }

    public void makeTransactionalAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeTransactionalAll(collection);
    }

    public void makeNontransactional(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeNontransactional(o);
    }

    public void makeNontransactionalAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        makeNontransactionalAll(objects);
    }

    public void makeNontransactionalAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().makeNontransactionalAll(collection);
    }

    public void retrieve(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().retrieve(o);
    }

    public void retrieveAll(Collection collection) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().retrieveAll(collection);
    }

    public void retrieveAll(Object[] objects) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().retrieveAll(objects);
    }

	public void retrieveAll(Collection collection, boolean b) {
		if (managedConnection == null) {
			throw new JDOException(NO_ACTIVE_CONNECTION);
		}
		managedConnection.getProxyManager().retrieveAll(collection, b);
	}

	public void retrieveAll(Object[] objects, boolean b) {
		if (managedConnection == null) {
			throw new JDOException(NO_ACTIVE_CONNECTION);
		}
		managedConnection.getProxyManager().retrieveAll(objects, b);
	}

    public void setUserObject(Object o) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().setUserObject(o);
    }

    public Object getUserObject() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getUserObject();
    }

    public PersistenceManagerFactory getPersistenceManagerFactory() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getPersistenceManagerFactory();
    }

    public Class getObjectIdClass(Class aClass) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getObjectIdClass(aClass);
    }

    public void setMultithreaded(boolean b) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().setMultithreaded(b);
    }

    public boolean getMultithreaded() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getMultithreaded();
    }

    public void setIgnoreCache(boolean b) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().setIgnoreCache(b);
    }

    public boolean getIgnoreCache() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getIgnoreCache();
    }
    
    public void addLifecycleListener(LifecycleListener arg0, Class[] arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().addLifecycleListener(arg0, arg1);
    }
    public Object attachCopy(Object arg0, boolean arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().attachCopy(arg0, arg1);
    }
    public Collection attachCopyAll(Collection arg0, boolean arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().attachCopyAll(arg0, arg1);
    }
    public Object[] attachCopyAll(Object[] arg0, boolean arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().attachCopyAll(arg0, arg1);
    }
    public Object detachCopy(Object arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().detachCopy(arg0);
    }
    public Collection detachCopyAll(Collection arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().detachCopyAll(arg0);
    }
    public Object[] detachCopyAll(Object[] arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().detachCopyAll(arg0);
    }
    public void flush() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().flush();
    }
    public void checkConsistency() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().checkConsistency();
    }
    public JDOConnection getDataStoreConnection() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getDataStoreConnection();
    }
    public Extent getExtent(Class arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getExtent(arg0);
    }
    public FetchPlan getFetchPlan() {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().getFetchPlan();
    }
    public Query newNamedQuery(Class arg0, String arg1) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        return managedConnection.getProxyManager().newNamedQuery(arg0, arg1);
    }
    public void refreshAll(JDOException arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().refreshAll(arg0);
    }
    public void removeLifecycleListener(LifecycleListener arg0) {
        if (managedConnection == null) {
            throw new JDOException(NO_ACTIVE_CONNECTION);
        }
        managedConnection.getProxyManager().removeLifecycleListener(arg0);
    }
}
