/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.mapper.lib;

import org.objectweb.perseus.persistence.api.StorageManager;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.NoDSIPersistenceException;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PExceptionNoDSI;
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.mapper.api.JormFactory;

import javax.jdo.InstanceCallbacks;
import java.util.Map;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This class is an implementation of the StorageManager interface based on
 * Jorm.The single hypothesis concerns the architecture of the CacheEntry:
 * - the CacheEntry implements the PBinding interface
 * - the obj parameters implement State and PAceessor.
 *
 * @author S.Chassande-Barrioz
 */
public class JormStorageManager
        implements StorageManager, BindingController {

    public final static String JORM_FACTORY_BINDING = "jorm-factory";
    public final static String LOGGER_NAME = "org.objectweb.jorm.storageManager";

    protected JormFactory jormFactory = null;

    protected Logger logger = null;

	/**
	 * listes for each working set the persistent object removed in other
	 * working set. When a working set want to read a persistent object from
	 * the data support, a prefetch buffer can be used. But the prefetch buffer
	 * can contains a removed object.
	 */
	protected Map ws2removedpo = new HashMap();

    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
        return new String[] { JORM_FACTORY_BINDING };
    }

    public Object lookupFc(String s) {
		if (JORM_FACTORY_BINDING.equals(s)) {
            return jormFactory;
        }
        return null;
    }

    public void bindFc(String s, Object o) {
        if ("logger".equals(s)) {
			logger = (Logger) o;
		} else if (JORM_FACTORY_BINDING.equals(s)) {
            jormFactory = (JormFactory) o;
        }
    }

    public void unbindFc(String s) {
		if (JORM_FACTORY_BINDING.equals(s)) {
            jormFactory = null;
        }
    }


	//IMPLEMENTATION OF THE StorageManager INTERFACE //
	//-----------------------------------------------//

    public Object export(ConnectionHolder context, Object obj) throws PersistenceException {
        PBinding pb = (PBinding) obj;
        try {
            if (pb.getStatus() == PBinding.LIFECYCLE_DELTOWRITE) {
                pb.write(context, (PAccessor) obj);
            }
            return pb.export(context);
        } catch (PException e) {
            throw new PersistenceException(e);
        }
    }

    public Object export(ConnectionHolder context, Object obj, Object hints) throws PersistenceException {
		PBinding pb = (PBinding) obj;
        try {
            if (pb.getStatus() == PBinding.LIFECYCLE_DELTOWRITE) {
                pb.write(context, (PAccessor) obj);
            }
            return pb.export(context, hints);
        } catch (PException e) {
            throw new PersistenceException(e);
        }
    }

    public void unexport(ConnectionHolder context, Object oid) throws PersistenceException {
        try {
            ((PName) oid).unexport(context);
        } catch (PException e) {
            throw new PersistenceException(e);
        }
		synchronized(ws2removedpo) {
			Iterator it = ws2removedpo.values().iterator();
			while(it.hasNext()) {
				HashSet removedpos = (HashSet) it.next();
				removedpos.add(oid);
			}
		}
    }

    public void unexport(ConnectionHolder context, Object oid, Object hints) throws PersistenceException {
        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "unexport ctx=" + context + " / oid=" + oid);
        try {
            ((PName) oid).unexport(context, hints);
        } catch (PException e) {
            throw new PersistenceException(e);
        }
		synchronized(ws2removedpo) {
			Iterator it = ws2removedpo.values().iterator();
			while(it.hasNext()) {
				HashSet removedpos = (HashSet) it.next();
				removedpos.add(oid);
			}
		}
    }

    public void read(ConnectionHolder context, Object oid, State obj) throws PersistenceException {
        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "read ctx=" + context
                    + " / oid=" + oid
                    + " / obj.class=" + obj.getClass().getName());
		PBinding pb = (PBinding) obj.getCacheEntry();
		if (pb.getStatus() == PBinding.LIFECYCLE_NEWTOWRITE) {
			if (logger.isLoggable(BasicLevel.DEBUG))
	            logger.log(BasicLevel.DEBUG, "Do not read an exported object");
		    return;
		}
        try {
            pb.read(context, (PAccessor) obj);
	        if (pb instanceof InstanceCallbacks) {
		        ((InstanceCallbacks) pb).jdoPostLoad();
	        }
        } catch (PExceptionNoDSI e) {
			throw new NoDSIPersistenceException(e);
		} catch (PException e) {
            throw new PersistenceException(e);
        }
    }

    public void read(WorkingSet ws, Object oid, State obj) throws PersistenceException {
        Object conn = ws.getConnectionHolder().getCHConnectionForRead();
		if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "read conn=" + conn
                    + " / oid=" + oid
                    + " / obj.class=" + obj.getClass().getName()
                    + " / tx=" + ws);
		PBinding pb = (PBinding) obj.getCacheEntry();
		if (pb.getStatus() == PBinding.LIFECYCLE_NEWTOWRITE) {
			if (logger.isLoggable(BasicLevel.DEBUG))
	            logger.log(BasicLevel.DEBUG, "Do not read an exported object");
		    return;
		}
		Object ctx = ws;
		//check if the persistent object has not been removed in another context.
		//Indeed in case of Jorm uses prefetch buffers, the persistent will be
		// loaded even if it has been removed from the data base.
		synchronized(ws2removedpo) {
			HashSet removedpos = (HashSet) ws2removedpo.get(ws);
			if (removedpos != null && removedpos.contains(oid)) {
				ctx = ws.getConnectionHolder();
			}
		}
        try {
            pb.read(conn, (PAccessor) obj, ctx);
	        if (pb instanceof InstanceCallbacks) {
		        ((InstanceCallbacks) pb).jdoPostLoad();
	        }
		} catch (PExceptionNoDSI e) {
			throw new NoDSIPersistenceException(e);
        } catch (PException e) {
            throw new PersistenceException(e);
        }
    }

    public void write(ConnectionHolder context, Object oid, State obj) throws PersistenceException {
		PBinding pb = (PBinding) obj.getCacheEntry();
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "write ctx=" + context
                    + " / oid=" + oid
                    + " / obj.class=" + obj.getClass().getName()
                    + " / binding.status=" + pb.getStatus());
        }
        int s = pb.getStatus();
        if (s == PBinding.LIFECYCLE_ACTIVEFORIO 
                || s == PBinding.LIFECYCLE_NEWTOWRITE) {
			if (pb instanceof InstanceCallbacks) {
				((InstanceCallbacks) pb).jdoPreStore();
			}
        }
        if (s == PBinding.LIFECYCLE_ACTIVEFORIO 
                || s == PBinding.LIFECYCLE_NEWTOWRITE
                || s == PBinding.LIFECYCLE_DELTOWRITE) {
			try {
				pb.write(context, (PAccessor) obj);
			} catch (PExceptionNoDSI e) {
				throw new NoDSIPersistenceException(e);
			} catch (PException e) {
				throw new PersistenceException(e);
			}
		}
    }

	/**
	 * registered the begining of a working set
	 */
	public void beginWS(WorkingSet ws) {
		synchronized(ws2removedpo) {
			ws2removedpo.put(ws, new HashSet());
		}
	}

	/**
	 * forget the working set
	 */
	public void endWS(WorkingSet ws) {
		synchronized(ws2removedpo) {
			ws2removedpo.remove(ws);
		}
	}

}
