/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.mim.api;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.query.api.QueryDefinition;

import java.util.Collection;
import java.util.Map;
import java.util.Properties;


/**
 * It defines a home for a persitent class. The home groups common fields and 
 * methods.
 * 
 * @author S.Chassande-Barrioz
 */
public interface SpeedoHome extends PClassMapping, Loggable {

    String getProjectName();

    /**
     * Indicates if the class has been marked as detacheable
     * @return
     */
    boolean isDetachable();
    
    /**
     * Indicates if the state of persistent instance managed by this home, are 
     * kept between two transaction.
     * @return true means the state is kept between transaction.
     */
    boolean isCacheable();
    
    /**
     * Permit to Indicate if the state of persistent instance managed by this 
     * home, are kept between two transaction.
     */
    void setCacheable(boolean v);
    
    /**
	 * The versioning strategy.
	 * @see SpeeedoVersion.TIMESTAMP STATE_COMPARISON VERSION_NUMBER NO_VERSION
	 */
	byte getVersioningStrategy();

	/**
	 * @return the PersistenceManagerFactory representing the data support
	 * where the proxy is persistent.
	 */
	ProxyManagerFactory getProxyManagerFactory();

	/**
	 * It assignes a PersistenceManagerFactory
	 * @param _pmf is the PersistenceManagerFactory representing the data 
	 * support where the proxy is persistent.
	 */
	void setProxyManagerFactory(ProxyManagerFactory _pmf);


	/**
	 * @return the TransactionalPersistenceManager which manages the
	 * concurrency, loading, caching ... of the proxies.
	 */
	TransactionalPersistenceManager getTransactionalPersistenceManager();

	/**
	 * It assignes a TransactionalPersistenceManager
	 * @param _tpm is the TransactionalPersistenceManager which manage the
	 * concurrency, loading, ... of the proxy.
	 */
	void setTransactionalPersistenceManager(TransactionalPersistenceManager _tpm);

	/**
	 * Notifies the transactional persistency manager of a read intention for
	 * the given persistent instance.
	 * @param sp is the SpeedoProxy which the read access is requested.
	 * @param fields the ids of the fields that may be accessed by the caller
	 * of this method. If the i-th bit of 'fields' is set to 1, then the i-th
	 * field of the given speedo accessor may be accessed by the caller of this
	 * method.
	 */
	SpeedoAccessor readIntention(SpeedoProxy sp, long[] fields);

	/**
	 * Notifies the transactional persistency manager of a write intention for
	 * the given persistent instance.
	 *
	 * @param sp is the SpeedoProxy which the write access is requested.
	 * @param fields the ids of the fields that may be accessed by the caller
	 * of this method. If the i-th bit of 'fields' is set to 1, then the i-th
	 * field of the given speedo accessor may be accessed by the caller of this
	 * method.
	 */
	SpeedoAccessor writeIntention(SpeedoProxy sp, long[] fields);

	SpeedoAccessor getSpeedoAccessor(SpeedoProxy sp);

	SpeedoProxy detachCopy(SpeedoProxy sp, 
	        ProxyManager pm, 
	        Map map, 
	        Object clone, 
	        Collection fgHints) ;

	void attachCopy(SpeedoProxy sp, 
		ProxyManager pm, 
		Map map, 
		Object clone, 
		SpeedoAccessor sa, 
		boolean makeTransactional) ;
	
	Collection fgGetNestedFetchGroups(String fgName); 

	/**
	 * @param fgName is a fetch group name
	 * @return true is the fetch group is defined, otherwise false.
	 */
	boolean fgIsDefined(String fgName);

	Collection fgGetFieldsToLoad(String fgName);

	Collection fgGetFieldsToLoad(String fgName, boolean onlyDirectRef);

	Properties getClassProperties();
	
	/**
	 * Defines a named query
	 * @param name is the name identifying the query
	 * @param query is the definition of the query
	 * @return the old query definition corresponding to the name, if it
	 * existed one.
	 */
	QueryDefinition addNamedQuery(String name, QueryDefinition query);

	/**
	 * Remove a named query
	 * @param name is the name of the query definition to remove
	 * @return the removed query definition (null if not found)
	 */
	QueryDefinition removeNamedQuery(String name);

	/**
	 * Retrieves a query defintion from its name. If the name is not known, the
	 * null value is returned.
	 * @param name is the name the wanted query definition
	 */
	QueryDefinition getNamedQuery(String name);
	
		
	void initSH();
	
}
