/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.pm.lib;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Properties;

import javax.jdo.DataStoreCache;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.speedo.Speedo;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.mapper.api.ConnectionSpecFactory;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.pm.api.ProxyManagerSwitch;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * The SpeepoProxyManagerFactory is a factory of ProxyManager. This primitive
 * component exports the ProxyManagerFactory interface, permitting to manage the
 *  proxy manager allocations. Due to a high cost of ProxyManager instances
 * allocation, the unsused ProxyManager are pooled. For this reason, the
 * SpeedoProxyManagerFactory uses a Pool of SpeedoProxyManager. The
 * SpeedoProxyManagerFactory manages also the attachement of ProxyManager to the
 * current thread through a ProxyManagerSwitch.
 *
 * @see org.objectweb.speedo.pm.api.ProxyManager
 * @see org.objectweb.speedo.pm.api.ProxyManagerSwitch
 * @see org.objectweb.speedo.pm.lib.SpeedoProxyManager
 * @see org.objectweb.speedo.pm.lib.SpeedoProxyManagerFactory
 * @see org.objectweb.speedo.pm.lib.SpeedoProxyManagerInstanciator
 * @see org.objectweb.speedo.pm.lib.SpeedoProxyManagerSwitch
 * 
 * @author S.Chassande-Barrioz
 */
public class SpeedoProxyManagerFactory
        implements ProxyManagerFactory,
		PersistenceManagerFactory,
		BindingController {

    public final static String MAPPER_BINDING = "mapper";
    public final static String PM_POOL_BINDING = "proxy-manager-pool";
    public final static String PROXY_MANAGER_SWITCH_BINDING = "proxy-manager-switch";
    public final static String COMPONENT_BINDING = "component";
    private final static String optionArray[] = {
    		SpeedoProperties.JDO_OPTION_TRANSIENT_TRANSACTIONAL, 
			//not yet supportded JDO_OPTION_NON_TRANSACTIONAL_READ, 
			//not yet supportded ed JDO_OPTION_NON_TRANSACTIONAL_WRITE, 
    		SpeedoProperties.JDO_OPTION_RETAIN_VALUES, 
			//not yet supportded JDO_OPTION_RESTORE_VALUES, 
    		SpeedoProperties.JDO_OPTION_OPTIMISTIC, 
    		SpeedoProperties.JDO_OPTION_APPLICATION_IDENTITY, 
    		SpeedoProperties.JDO_OPTION_DATASTORE_IDENTITY, 
    		SpeedoProperties.JDO_OPTION_NON_DURABLE_IDENTITY,
    		SpeedoProperties.JDO_OPTION_ARRAY_LIST,
    		SpeedoProperties.JDO_OPTION_HASH_MAP,
    		SpeedoProperties.JDO_OPTION_HASH_TABLE, 
    		SpeedoProperties.JDO_OPTION_LINKED_LIST, 
    		SpeedoProperties.JDO_OPTION_TREE_MAP,
    		SpeedoProperties.JDO_OPTION_TREE_SET, 
    		SpeedoProperties.JDO_OPTION_VECTOR,
    		SpeedoProperties.JDO_OPTION_ARRAY,
    		SpeedoProperties.JDO_OPTION_MAP,
    		SpeedoProperties.JDO_OPTION_LIST,			
    		SpeedoProperties.JDO_OPTION_NULL_COLLECTION, 
    		SpeedoProperties.JDO_QUERY_JDOQL
    };

    private final static int GETMAXPOOL = 100;

    /**
     * Pool of PersistenceManager
     */
    private Pool managedPM = null;

    /**
     * Manages the log's traces
     */
    private Logger logger = null;

    /**
     * JORM object encapsulating the datastore
     */
    private PMapper mapper = null;

    private Properties connectionProperties = null;

    private ProxyManagerSwitch pms = null;

    private Object _this = null;

    private ConnectionSpecFactory csf = null;

	private boolean started = false;


    public SpeedoProxyManagerFactory() {
        connectionProperties = new Properties();
    }

	private ProxyManagerFactory getThis() {
		if (_this == null) {
			throw new JDODataStoreException(
					"Cannot provides a PersistenceManager without a reference to the component");
		} else if (_this instanceof Component) {
			try {
				_this = ((Component)_this).getFcInterface("proxy-manager-factory");
			} catch (NoSuchInterfaceException e) {
				throw new JDODataStoreException(
						"Cannot provides a PersistenceManager ", new Exception[]{e});
			}
		}
		return (ProxyManagerFactory) _this;
	}

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public String[] listFc() {
        return new String[] {
            PM_POOL_BINDING,
            MAPPER_BINDING,
            PROXY_MANAGER_SWITCH_BINDING,
        };
    }

    public Object lookupFc(String s) {
        if (PM_POOL_BINDING.equals(s))
            return managedPM;
        else if (MAPPER_BINDING.equals(s))
            return mapper;
        else if (PROXY_MANAGER_SWITCH_BINDING.equals(s))
            return pms;
        else
            return null;
    }

    public void bindFc(String s, Object o) {
        if ("logger".equals(s)) {
            logger = (Logger) o;
        } else if (PM_POOL_BINDING.equals(s)) {
            managedPM = (Pool) o;
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = (PMapper) o;
        } else if (PROXY_MANAGER_SWITCH_BINDING.equals(s)) {
            pms = (ProxyManagerSwitch) o;
        } else if (COMPONENT_BINDING.equals(s)) {
            _this = o;
        }
    }

    public void unbindFc(String s) {
        if (PM_POOL_BINDING.equals(s)) {
            managedPM = null;
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = null;
        } else if (PROXY_MANAGER_SWITCH_BINDING.equals(s)) {
            pms = null;
        }
    }

    // IMPLEMENTATION OF THE DataStoreCache INTERFACE //
    //------------------------------------------------//
    public void evict(Object arg0) {
    }
    public void evictAll(Class arg0, boolean arg1) {
    }
    public void evictAll(Collection arg0) {
    }
    public void evictAll(Object[] arg0) {
    }
    public void pin(Object arg0) {
    }
    public void pinAll(Class arg0, boolean arg1) {
    }
    public void pinAll(Collection arg0) {
    }
    public void pinAll(Object[] arg0) {
    }
    public void unpin(Object arg0) {
    }
    public void unpinAll(Class arg0, boolean arg1) {
    }
    public void unpinAll(Collection arg0) {
    }
    public void unpinAll(Object[] arg0) {
    }

    // IMPLEMENTATION OF THE ProxyManagerFactory INTERFACE //
    //-----------------------------------------------------//

	public ProxyManager lookup() {
		return pms.lookup(getThis());
	}

	public void bindPM2Thread(ProxyManager pm) {
		if (logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "Bind a pm to the thread: " + pm);
		}
		pms.bind(pm);
	}

	public void unbindPM() {
		if (logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "Unbind the current PM from the thread");
		}
		pms.unbind(getThis());
	}

	public void proxyManagerClosed(ProxyManager pr) {
		try {
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "Unbind a PM from the thread: " + pr);
			}
			pms.unbind(pr);
			managedPM.releaseResource(pr);
		} catch (Exception e) {
			throw new JDOFatalInternalException(
					"Impossible to release this PersistenceManager in the pool.",
					new Exception[]{ExceptionHelper.getNested(e)});
		}
	}


    // IMPLEMENTATION OF THE PersistentManagerFactory INTERFACE //
    //----------------------------------------------------------//

    /** Get an instance of PersistenceManager from this factory. The instance has
     * default values for options.
     * Invokes <code>init</code> at the first call.
     *
     * @return a PersistenceManager instance with default options.
     */
    public PersistenceManager getPersistenceManager() {
        return getPersistenceManager(null);
    }

	public void close() {
		//TODO: implements the close method on the PMF
	}

    /** Get an instance of PersistenceManager from this factory.  The instance has
     * default values for options.  The parameters userid and password are used
     * when obtaining datastore connections from the connection pool.
     * Invokes <code>init</code> at the first call.
     *
     * @param userid the userid for the connection
     * @param password the password for the connection
     * @return a PersistenceManager instance with default options.
     */
    public PersistenceManager getPersistenceManager(String userid, String password) {
        Object cs = null;
        if (csf == null) {
            String cn = connectionProperties.getProperty(Speedo.CONNECTION_SPEC_FACTORY);
            if (cn != null && cn.length() > 0) {
                try {
                    csf = (ConnectionSpecFactory) Class.forName(cn).newInstance();
                    cs = csf.getConnectionSpec(this, userid, password);
                } catch (Exception e) {
                    logger.log(BasicLevel.ERROR,
                        "Impossible to instanciate the ConnectionSpecFactory: "
                        + cn, e);
                }
            }
        } else {
            cs = csf.getConnectionSpec(this, userid, password);
        }
        return getPersistenceManager(cs);
    }

    public synchronized PersistenceManager getPersistenceManager(Object cs) {
        started = true;        
		ProxyManager pm = pms.lookup(getThis());
        if (pm != null && cs != null) {
            pm.getSemaphore().P();
            try {
                if (!pm.isClosed()) {
                    logger.log(BasicLevel.INFO, "reuse the same PersistenceManager (Imbricated)");
                    pm.addUse();
                    return pm;
                }
            } finally {
                pm.getSemaphore().V();
            }
        }
        
        try {
            // Gets a PM from the pool
            Object pr = managedPM.getResource(null);
			logger.log(BasicLevel.DEBUG, "get a persistenceManager from the pool");
            Component ci = ((Interface) pr).getFcItfOwner();
            pm = (ProxyManager) ci.getFcInterface("proxy-manager");
            pm.open(cs);
            bindPM2Thread(pm);
            pm.addUse();
            return pm;
        } catch (Exception e) {
            throw new JDODataStoreException(
                    "Cannot provides a PersistenceManager ", new Exception[]{e});
        }
    }

    /* (non-Javadoc)
     * @see javax.jdo.PersistenceManagerFactory#getDataStoreCache()
     */
    public DataStoreCache getDataStoreCache() {
        // TODO Auto-generated method stub
        return null;
    }
    /* (non-Javadoc)
     * @see javax.jdo.PersistenceManagerFactory#isClosed()
     */
    public boolean isClosed() {
        return false;
    }
    public String getMapping() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_MAPPING, "no user name defined");
    }
    public void setMapping(String arg0) {
        connectionProperties.setProperty(SpeedoProperties.JDO_OPTION_MAPPING, arg0);
    }

    /** Set the user name for the data store connection.
     * @param userName the user name for the data store connection.
     */
    public void setConnectionUserName(String userName) {
        connectionProperties.setProperty(SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME, userName);
    }

    /** Get the user name for the data store connection.
     * @return the user name for the data store connection.
     */
    public String getConnectionUserName() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME, "no user name defined");
    }

    /** Set the password for the data store connection.
     * @param password the password for the data store connection.
     */
    public void setConnectionPassword(String password) {
        connectionProperties.setProperty(SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD, password);
    }

    /** Get the password for the data store connection.
     * @return  password the password for the data store connection.
     */
    protected String getConnectionPassword() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD);
    }

    /** Set the URL for the data store connection.
     * @param URL the URL for the data store connection.
     */
    public void setConnectionURL(String URL) {
        connectionProperties.setProperty(SpeedoProperties.JDO_OPTION_CONNECTION_URL, URL);
    }

    /** Get the URL for the data store connection.
     * @return the URL for the data store connection.
     */
    public String getConnectionURL() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_URL);
    }

    /** Set the driver name for the data store connection.
     * @param driverName the driver name for the data store connection.
     */
    public void setConnectionDriverName(String driverName) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME, driverName);
    }

    /** Get the driver name for the data store connection.
     * @return the driver name for the data store connection.
     */
    public String getConnectionDriverName() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME);
    }

    /** Set the name for the data store connection factory.
     * @param connectionFactoryName the name of the data store connection factory.
     */
    public void setConnectionFactoryName(String connectionFactoryName) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY_NAME, connectionFactoryName);
    }

    /** Get the name for the data store connection factory.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactoryName() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY_NAME);
    }

    /** Set the data store connection factory.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionFactory the data store connection factory.
     */
    public void setConnectionFactory(Object connectionFactory) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        throw new JDOUnsupportedOptionException("Cannot set a Connection Factory, use <> instead");
    }

    /** Get the data store connection factory.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory() {
        return mapper.getConnectionFactory();
    }

    /** Set the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @param connectionFactoryName the name of the data store connection factory.
     */
    public void setConnectionFactory2Name(String connectionFactoryName) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY2_NAME, connectionFactoryName);
    }

    /** Get the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactory2Name() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY2_NAME);
    }

    /** Set the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionfactory the data store connection factory.
     */
    public void setConnectionFactory2(Object connectionfactory) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        throw new JDOUnsupportedOptionException(
                "Cannot set a second Connection Factory");
    }

    /** Get the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory2() {
        return mapper.getConnectionFactory();
    }

    /** Set the default Multithreaded setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default Multithreaded setting.
     */
    public void setMultithreaded(boolean flag) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_MULTITREADED, new Boolean(flag).toString());
    }

    /** Get the default Multithreaded setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default Multithreaded setting.
     */
    public boolean getMultithreaded() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_MULTITREADED, "").equals("true");
    }

    /** Set the default Optimistic setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default Optimistic setting.
     */
    public void setOptimistic(boolean flag) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_OPTIMISTIC, new Boolean(flag).toString());
    }

    /** Get the default Optimistic setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default Optimistic setting.
     */
    public boolean getOptimistic() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_OPTIMISTIC,"").equals("true");
    }

    /** Set the default RetainValues setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default RetainValues setting.
     */
    public void setRetainValues(boolean flag) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_RETAIN_VALUES, new Boolean(flag).toString());
    }

    /** Get the default RetainValues setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default RetainValues setting.
     */
    public boolean getRetainValues() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_RETAIN_VALUES).equals("true");
    }

    /** Set the default NontransactionalRead setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default NontransactionalRead setting.
     */
    public void setNontransactionalRead(boolean flag) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_READ, new Boolean(flag).toString());
    }

    /** Get the default NontransactionalRead setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default NontransactionalRead setting.
     */
    public boolean getNontransactionalRead() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_READ).equals("true");
    }

    /** Set the default NontransactionalWrite setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default NontransactionalWrite setting.
     */
    public void setNontransactionalWrite(boolean flag) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_WRITE, new Boolean(flag).toString());
    }

    /** Get the default NontransactionalWrite setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default NontransactionalWrite setting.
     */
    public boolean getNontransactionalWrite() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_WRITE).equals("true");
    }

    /** Set the default IgnoreCache setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default IgnoreCache setting.
     */
    public void setIgnoreCache(boolean flag) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(
        		SpeedoProperties.JDO_OPTION_IGNORE_CACHE, new Boolean(flag).toString());
    }

    /** Get the default IgnoreCache setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default IngoreCache setting.
     */
    public boolean getIgnoreCache() {
        return connectionProperties.getProperty(SpeedoProperties.JDO_OPTION_IGNORE_CACHE).equals("true");
    }

    /** Get the MaxPool setting for the PersistenceManager
     * pool for this factory.
     * @return the MaxPool setting.
     */
    public int getMaxPool() {
		String v = connectionProperties.getProperty(SpeedoProperties.PM_POOL_MAX);
		if (v != null && v.length()>0) {
			return Integer.parseInt(v);
		} else {
			return GETMAXPOOL;
		}
    }

    /** Set the MaxPool setting for the PersistenceManager
     * pool for this factory.
     * @param maxPool the MaxPool setting.
     */
    public void setMaxPool(int maxPool) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(SpeedoProperties.PM_POOL_MAX, Integer.toString(maxPool));
    }

    /** Get the MinPool setting for the PersistenceManager
     * pool for this factory.
     * @return the MinPool setting.
     */
    public int getMinPool() {
        return Integer.parseInt(connectionProperties.getProperty(SpeedoProperties.PM_POOL_MIN));
    }

    /** Set the MinPool setting for the PersistenceManager
     * pool for this factory.
     * @param minPool the MinPool setting.
     */
    public void setMinPool(int minPool) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(SpeedoProperties.PM_POOL_MIN, Integer.toString(minPool));
    }

    /** Get the MsWait setting for the PersistenceManager
     * pool for this factory.
     * @return the MsWait setting.
     */
    public int getMsWait() {
        return Integer.parseInt(connectionProperties.getProperty(SpeedoProperties.PM_POOL_TIMEOUT));
    }

    /** Set the MsWait setting for the PersistenceManager
     * pool for this factory.
     * @param msWait the MsWait setting.
     */
    public void setMsWait(int msWait) {
		if (started) {
			throw new JDOUserException("It is forbidden to modify property after a getPersistenceManager() call");
		}
        connectionProperties.setProperty(SpeedoProperties.PM_POOL_TIMEOUT, Integer.toString(msWait));
    }

    /** Return non-configurable properties of this PersistenceManagerFactory.
     * Properties with keys VendorName and VersionNumber are required.  Other
     * keys are optional.
     * @return the non-configurable properties of this
     * PersistenceManagerFactory.
     */
    public Properties getProperties() {
        return connectionProperties;
    }

    /** The application can determine from the results of this
     * method which optional features, and which query languages
     * are supported by the JDO implementation.
     * <P>Each supported JDO optional feature is represented by a
     * String with one of the following values:
     *
     * <P>javax.jdo.option.TransientTransactional
     * <P>javax.jdo.option.NontransactionalRead
     * <P>javax.jdo.option.NontransactionalWrite
     * <P>javax.jdo.option.RetainValues
     * <P>javax.jdo.option.Optimistic
     * <P>javax.jdo.option.ApplicationIdentity
     * <P>javax.jdo.option.DatastoreIdentity
     * <P>javax.jdo.option.NonDatastoreIdentity
     * <P>javax.jdo.option.ArrayList
     * <P>javax.jdo.option.HashMap
     * <P>javax.jdo.option.Hashtable
     * <P>javax.jdo.option.LinkedList
     * <P>javax.jdo.option.TreeMap
     * <P>javax.jdo.option.TreeSet
     * <P>javax.jdo.option.Vector
     * <P>javax.jdo.option.Map
     * <P>javax.jdo.option.List
     * <P>javax.jdo.option.Array
     * <P>javax.jdo.option.NullCollection
     *
     *<P>The standard JDO query language is represented by a String:
     *<P>javax.jdo.query.JDOQL
     * @return the List of String representing the supported Options
     */
    public Collection supportedOptions() {
    	return Collections.unmodifiableList(Arrays.asList(optionArray));    	
    }

    public void setRestoreValues(boolean b) {
    }

    public boolean getRestoreValues() {
        return false;
    }

    // IMPLEMENTATION OF THE PoolProvider INTERFACE //
    //----------------------------------------------//

    /**
     * @return the owned <code>Pool</code>
     */
    public Pool getPool() {
        return managedPM;
    }
}
