/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S. Chassande-Barrioz
 *
 */

package org.objectweb.speedo.query.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.PersistenceException;

import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOUserException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * is the result of a JDO query. The result is exported as a
 * java.util.Collection of persistent instances. This implementation supports
 * to have a single Iterator.
 *
 * @author S.Chassande-Barrioz
 */
public class QueryResultCollection implements Collection {

    /**
	 * The inner medor TupleCollection
	 */
	private TupleCollection tc = null;

	/**
	 * The proxy manager representing the current execution context.
	 */
	private PersistenceManager pm = null;

	/**
	 * The underlying connection used to evalue the query. These connections
	 * will be closed when the result will be no more used.
	 */
	private Object[] conns = null;

	/**
	 * is the index in the tuple structure (of the inner Medor TupleCollection)
	 * of the PName of the expected persistent instance.
	 */
	private int pnIndex = -1;


	/**
	 * if all values are loaded, this field represents the result of the query.
	 */
	private List values = null;

	/**
	 * Is the current iteraor over the query result.
	 */
	private QueryIterator iterator;

	private Logger logger;
	private boolean debug;


    public QueryResultCollection(TupleCollection tc,
                     PersistenceManager pm,
                     Object[] conns,
					 boolean fixablePNIndex,
                     Logger logger) throws MedorException {
        this.tc = tc;
        this.pm = pm;
        this.conns = conns;
		this.logger = logger;
		debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
		if (fixablePNIndex) {
        	this.pnIndex = getPNIndex();
			if (debug) {
				logger.log(BasicLevel.DEBUG, "index of the pname: " + pnIndex);
			}
		}
		if (pm == null || pm.isClosed() || ((Transaction) pm.currentTransaction()).getStatus()
				== WorkingSet.CTX_CLOSED) {
			throw new JDOUserException(
					"Impossible to use a query result without opened persistence manager");
		}
    }

    /**
	 * Close the connection to the data store. The query result will be no
	 * more usable after.
	 */
	public void close() {
        pm = null;
		if (conns != null) {
			for(int i=0; i<conns.length; i++) {
				if (conns[i] instanceof ConnectionHolder) {
					try {
						((ConnectionHolder) conns[i]).releaseCHConnection();
					} catch (PersistenceException e) {
					}
				}
			}
		}
        conns = null;
    }

    /**
     * Computes the query result.
     */
    private List getList() {
		if (pm == null) {
			throw new JDOUserException("Impossible to use a closed query result");
		}
		if (pm.isClosed() || ((Transaction) pm.currentTransaction()).getStatus()
				== WorkingSet.CTX_CLOSED) {
			throw new JDOUserException(
					"Impossible to use a query result without opened persistence manager");
		}
        if (values != null)
            return values;

        values = new ArrayList();
        try {
            if (!tc.isEmpty()) {
                tc.first();
                do {
					PName pn ;
					if (pnIndex == -1) {;
						pn = (PName) tc.getTuple().getObject(getPNIndex());
					} else {
						pn = (PName) tc.getTuple().getObject(pnIndex);
					}
					if (!pn.isNull()) {
						try {
							values.add(pm.getObjectById(pn, false));
						} catch (JDOUserException e) {
							if (debug) {
								logger.log(BasicLevel.DEBUG,
									"Ignore a deleted object, oid=" + pn, e);
							}
							// Do nothing, the persistent object does not exist
							// ==> ignore it and go to the next
						}
					}
                } while (tc.next());
            }
        } catch (MedorException me) {
            throw new JDOUnsupportedOptionException(
                    "Persistence manager problem to get result", new Exception[]{me});
        }
        return values;
    }

	private int getPNIndex() throws MedorException {
		return tc.getMetaData().getSize();
	}

    private class QueryIterator implements Iterator {

		private Object next;

        public QueryIterator() {
            try {
                tc.first();
				if (pm != null && !pm.isClosed() && !tc.isEmpty()) {
					calculateNext(true);
				} else {
					next = null;
				}
            } catch (MedorException e) {
				next = null;
            }
        }

		public void desactivate() {
			next = null;
		}
        public boolean hasNext() {
            return next != null;
        }

        public Object next() {
			if (next == null) {
				throw new NoSuchElementException("No such element");
			}
			Object res = next;
			calculateNext(false);
			return res;
        }

		private void calculateNext(boolean isfirst) {
			boolean first = isfirst;
			next = null;
			if (pm == null || pm.isClosed()) {
				return;
			}
			try {
				while(next == null && (first || tc.next())) {
					first = false;
					PName pn ;
					if (pnIndex == -1) {;
						pn = (PName) tc.getTuple().getObject(getPNIndex());
					} else {
						pn = (PName) tc.getTuple().getObject(pnIndex);
					}
					if (!pn.isNull()) {
						try {
							next = pm.getObjectById(pn, false);
						} catch (JDOUserException e) {
							if (debug) {
								logger.log(BasicLevel.DEBUG,
									"Ignore a deleted object, oid=" + pn, e);
							}
							// Do nothing, the persistent object does not exist
							// ==> ignore it and go to the next
						}
					}
				}
			} catch (MedorException e) {
				logger.log(BasicLevel.ERROR, "Impossible to fetch the pname", e);
				throw new NoSuchElementException(e.getMessage());
			}
		}

        public void remove() {
            throw new UnsupportedOperationException(
                    "This method is not supported in this implementation");
        }
    }


    // IMPLEMENTATION OF THE Collection INTERFACE //
    //--------------------------------------------//

    /**
     * Ensures that this collection contains the specified element
     * (optional operation).
     */
    public boolean add(Object o) {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Adds all of the elements in the specified collection to this collection
     * (optional operation).
     */
    public boolean addAll(Collection c) {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Removes all of the elements from this collection
     * (optional operation).
     */
    public void clear() {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Returns true if this collection contains the specified element.
     */
    public boolean contains(Object o) {
        return getList().contains(o);
    }

    /**
     * Returns true if this collection contains all of the elements in the
     * specified collection.
     */
    public boolean containsAll(Collection c) {
        getList();
        for (Iterator iter = c.iterator(); iter.hasNext();) {
            if (!values.contains(iter.next()))
                return false;
        }
        return true;
    }

    /**
     * Compares the specified object with this collection for equality.
     */
    public boolean equals(Object o) {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Returns the hash code value for this collection.
     */
    public int hashCode() {
        return getList().hashCode();
    }

    /**
     * Returns true if this collection contains no elements.
     */
    public boolean isEmpty() {
        try {
            return tc == null || tc.isEmpty();
        } catch (MedorException me) {
            throw new JDOUnsupportedOptionException(
                    "a MedorException has been catched",
                    new Exception[]{me});
        }
    }

    /**
     * Returns an iterator over the elements in this collection.
     */
    public Iterator iterator() {
		if (iterator != null) {
			iterator.desactivate();
		}
		iterator = new QueryIterator();
        return iterator;
    }

    /**
     * Removes a single instance of the specified element from this collection,
     * if it is present (optional operation).
     */
    public boolean remove(Object o) {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Removes all this collection's elements that are also contained in the
     * specified collection (optional operation).
     */
    public boolean removeAll(Collection c) {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Retains only the elements in this collection that are contained in the
     * specified collection (optional operation).
     */
    public boolean retainAll(Collection c) {
        throw new JDOUnsupportedOptionException(
                "This method is not supported in this implementation");
    }

    /**
     * Returns the number of elements in this collection.
     */
    public int size() {
        return getList().size();
    }

    /**
     * Returns an array containing all of the elements in this collection.
     */
    public Object[] toArray() {
        return getList().toArray();
    }

    /**
     * Returns an array containing all of the elements in this collection whose
     * runtime type is that of the specified array.
     */
    public Object[] toArray(Object[] a) {
        return getList().toArray(a);
    }
}
