/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.usercache.lib;

import org.objectweb.speedo.usercache.api.UserCacheView;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.Map;

/**
 * Is a basic UserCacheView implementation using WeakReference in order to 
 * not disturb the real cache of persistent object.
 * 
 * @author S.Chassande-Barrioz
 */
public class UserCacheViewImpl implements UserCacheView {

	/**
	 * This field is the cache. This is a Map which the key are Class objects 
	 * and the values are a map linking an user key to a persistent instance.
	 * 
	 * Map(Class persistentClass, Map(Object userkey, Object persistentObject))  
	 */
	protected Map cache;

	/**
	 * Queue used to register the weak references stored in the cache.
	 */
	protected ReferenceQueue queue;

	/**
	 * The entries stored in 'cache' are weak references to the real entries.
	 * Each weak entry also stores the user key and the class of the entry 
	 * (these id and clazz are used in clean to remove the entry from 'cache')
	 */
	private static class WeakCacheEntry extends WeakReference {

		/**
		 * The user key of persistent object
		 */
		final Object userkey;

		/**
		 * The class of the entry 
		 */
		final Class clazz;

		public WeakCacheEntry(Object userkey, Object pc, ReferenceQueue queue) {
			super(pc, queue);
			this.userkey = userkey;
			this.clazz = pc.getClass();
		}
	}

	/**
	 * 
	 */
	public UserCacheViewImpl() {
		cache = new HashMap();
		queue = new ReferenceQueue();
	}

	/**
	 * Removes all entries that have been garbage collected.
	 */
	public synchronized void clean() {
		WeakCacheEntry ref;
		do {
			ref = (WeakCacheEntry) queue.poll();
			if (ref != null) {
				Object oid = ref.userkey;
				Map key2wce = (Map) cache.get(ref.clazz);
				if (key2wce != null) {
					//Remove the entry for the persistent instance
					key2wce.remove(oid);
					if (key2wce.isEmpty()) {
						//No more instance for the class => remove the map 
						cache.remove(ref.clazz);
					}
				}
			}
		} while(ref != null);
	}

	public synchronized Object bind(Object id, Object obj) {
		Map key2wce = (Map) cache.get(obj.getClass());
		if (key2wce == null) { 
			//no instance of the class already bound
			key2wce = new HashMap();
			cache.put(obj.getClass(), key2wce);
		}
		//register the instance
		Object old = key2wce.put(id, new WeakCacheEntry(id, obj, queue)); 
		// auto cleaning after registration 
		clean();
		return old;
	}

	public synchronized Object rebind(Object oldkey, 
			Object newkey, Object obj) {
		Map key2wce = (Map) cache.get(obj.getClass());
		if (key2wce == null) { 
			//no instance of the class already bound
			key2wce = new HashMap();
			cache.put(obj.getClass(), key2wce);
		} else {
			//remove the old binding
			key2wce.remove(oldkey);
		}
		//register the instance with the new key
		Object old = key2wce.put(newkey, new WeakCacheEntry(newkey, obj, queue));
		// auto cleaning after registration 
		clean();
		return old;
	}


	public synchronized Object unbind(Class clazz, Object key) {
		Map key2wce = (Map) cache.get(clazz);
		if (key2wce == null) {
			return null;
		}
		WeakCacheEntry ref = (WeakCacheEntry) key2wce.get(key);
		if (ref == null) {
			return null;
		} else {
			Object obj = (Object) ref.get();
			if (obj == null) {
				//auto-cleaning the entry and maybe others
				clean();
			} else {
				key2wce.remove(key);
				if (key2wce.isEmpty()) {
					//No more instance for the class => remove the map 
					cache.remove(ref.clazz);
				}
			}
			return obj;
		}
	}
	
	public synchronized Object lookup(Class clazz, Object key) {
		Map key2wce = (Map) cache.get(clazz);
		if (key2wce == null) {
			return null;
		}
		WeakCacheEntry ref = (WeakCacheEntry) key2wce.get(key);
		if (ref == null) {
			return null;
		} else {
			Object obj = (Object) ref.get();
			if (obj == null) {
				//auto-cleaning the entry and maybe others
				clean();
			}
			return obj;
		}
	}
}
