/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo.wizard;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.editor.wizard.api.IStrategy;
import org.objectweb.jorm.editor.wizard.api.IWizardProperties;
import org.objectweb.jorm.editor.wizard.jdo.api.IJdoWizard;
import org.objectweb.jorm.editor.wizard.jdo.api.IJdoWizardParameter;
import org.objectweb.speedo.generation.api.SpeedoCompilerParameter;
import org.objectweb.speedo.api.SpeedoException;
import org.apache.tools.ant.types.DTDLocation;

import java.util.Properties;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.File;
import java.io.FileInputStream;



/**
     * The SpeedoWizard class embodies a functional object and a
     * presentation object:
     * - The functional object encapsulates the algorithm that
     * generates JORM files from JDO specifications. The SpeedoWizard
     * provides a public run() method to initialize and execute
     * this algorithm.
     * - The presentation object displays a JDO wizard, gets the
     * user data and calls the run() method on the IStrategy object
     * it references.
     *
     * The SpeedoWizard class implements the IStrategy interface
     * imported from the wizard library. More precisely, it implements
     * a specific run() method.
     *
     * @author C. Hamon
     */

public class SpeedoWizard implements IStrategy {

    /**
     * The SpeedoWizardCompiler object represents the functional
     * part of SpeedoWizard. It creates and runs the tasks that
     * are involved in the generation process of JORM files.
     */
    private SpeedoWizardCompiler speedoWizardCompiler = null;

    /**
     * presentation references a JdoWizard object. It represents
     * the graphical part of SpeedoWizard.
     */
    private IJdoWizard presentation = null;

    /**
     *  Loads and reads the wizard.properties file.
     */
    private Properties properties = null;

    /**
     * Builds a SpeedoWizard object.
	 * @param swc               a SpeedoWizardCompiler object,
     *        property file	    the wizard.properties file.
     */
    public SpeedoWizard(SpeedoWizardCompiler swc, String propertyFile) throws PException {
        this.speedoWizardCompiler = swc;
        if (properties != null) {
            throw new PException("SpeedoWizard already configured.");
        }
        String cn = null;
        try {
            File pf = new File(propertyFile);
            properties = new Properties();
            properties.load(new FileInputStream(pf));
            } catch (Exception e) {
                properties = null;
                throw new PException(e, "Problem while configuring the SpeedoWizard: unable to locate [" + propertyFile + "]");
            }
            try {
                cn = properties.getProperty(IWizardProperties.JDO_WIZARD, "");
                presentation = (IJdoWizard) Class.forName(cn).newInstance();
                presentation.setStrategy(this);
            } catch (InstantiationException e) {
                throw new PException(e, "Problem while instantiating " + cn);
            } catch (IllegalAccessException e) {
                throw new PException(e, "Problem while trying to access " + cn);
            } catch (ClassNotFoundException e) {
                throw new PException(e, "Problem while loading class " + cn);
            }
    }

    /**
     * Displays the graphical JDO wizard.
     */
    public void show(String projectDir) throws Exception {
        presentation.show(projectDir);
    }

    /**
     * Initializes the SpeedoCompilerParameter object with data
     * provided by the JdoWizardParameter object.
     *
     * The data contained in the SpeedoCompilerParameter object is
     * then passed down to the different tasks involved in the
     * generation of the JORM files.
     *
     * Notes:
     * - scp.projectName contains a JORM project name. If it is set
     * to null, then no implicit JORM mapping data will be created.
     * - scp.mapperName contains a mapper name (for example: rdb.mysql)
     * or a null value.
     * Note that the user cannot specify a mapper name without first
     * indicating the name of the project.
     * - scp.jdo contains a collection of jdo files designated by their
     * absolute path names minus the first file separator character.
     * Note that .jdo files with absolute path names cannot be passed
     * down to the JDO parser.
     * - scp.jdoDir is set to empty string.
     * - scp.input contains either the root directory under which
     * .class files can be found, or a jar file hosting these
     * .class files. Note that a relative path is enough to retrieve
     * the .class files since the persistent class names and their
     * package name are known once the .jdo files have been parsed.
     *  - scp.jorm contains a collection of existing .pd files with
     * absolute paths. Note that .pd files with absolute paths can
     * directly be passed down to the JORM parser. If there exists
     * both a JDO specification and a JORM specification for a given
     * Java class, then a comparison will be done against these two
     * specifications.
     *  - scp.pdoutput contains the root directory under which .pd
     * files will be generated. Note that a relative path is enough
     * to build the full output directory path for each .pd file.
     * Indeed, the persistent class names and their package name are
     * known once the .jdo files have been parsed.
     *
     * @see org.objectweb.speedo.generation.parser.JdoParser
     * @see org.objectweb.speedo.generation.enhancer.MetaDataAnalyzer
     * @see org.objectweb.speedo.generation.enhancer.EnhancerComponent
     * @see org.objectweb.speedo.wizard.generation.JormWriter
     */
    private void initCompilerParameter()  {
        SpeedoCompilerParameter scp = speedoWizardCompiler.getSpeedoCompilerParameter();
        IJdoWizardParameter jdoWizardParameter = presentation.getJdoWizardParameter();
        scp.projectName = jdoWizardParameter.getProjectName();
        scp.mapperName = jdoWizardParameter.getMapperName();
        // Used by JdoParser:
        //          for (Iterator it = scp.jdo.iterator(); it.hasNext();) {
        //          String xmlFileName = (String) it.next();
        //          parser.parse(scp.jdoDir + File.separator + xmlFile);
        // Iterates on .jdo file names and removes the first file separator character
        scp.jdo = new ArrayList();
        for (Iterator it = jdoWizardParameter.getJdo().iterator(); it.hasNext();) {
            String xmlFileName = (String) it.next();
            scp.jdo.add(xmlFileName.substring(1));
        }
        //scp.jdo = jdoWizardParameter.jdo;
        scp.jdoDir = "";
        // Used by MetaDataAnalyser and is then pass to EnhancerComponent:
        //          loadJavaClass(isSrcJar,
        //                        jdoclass.jdoPackage.name + '.' + jdoclass.name,
        //                        scp.input, false);    -- old version
        //                        scp.output, false);    -- last version
        scp.input = jdoWizardParameter.getBinaryDir();
        scp.output = jdoWizardParameter.getBinaryDir();
    /*  scp.isSrcJar = jdoWizardParameter.isJar; */
        // Used by JormWriter:
        scp.jormDir = jdoWizardParameter.getJormDir();
        // Used by JormWriter and is then pass to JormCompiler:
        //          jcp.inputFiles = scp.jorm;
        //          Collection c = jormcompiler.getCompilerParameter().inputFiles;
        //          jormcompiler.parseFiles(c)
        scp.jorm = jdoWizardParameter.getJorm();
        // Used by JormWriter and is then pass to JormCompiler:
        //          jcp.output = scp.pdoutput;
        //          jormcompiler.generateJormFiles(createdMOs);
        scp.pdoutput = jdoWizardParameter.getOutputDir();
//        scp.jormclasspath ; initialized in Antjdo2jormGUI
        if (scp.dtdLocations.size() == 0) {
            DTDLocation dl = new DTDLocation();
            dl.setPublicId("-//ObjectWeb Consortium//DTD JORM DEFINITIONS 2.0//EN");
            dl.setLocation("jorm2.dtd");
            scp.dtdLocations.add(dl);
            dl = new DTDLocation();
            dl.setPublicId("-//Sun Microsystems, Inc.//DTD Java Data Objects Metadata 1.0//EN");
            dl.setLocation("jdo.dtd");
            scp.dtdLocations.add(dl);
        }
    }

    /**
     * Runs the functional part of the wizard
     * @return a boolean that indicates the success or the failure of the process
     */
    public boolean run() throws SpeedoException {
        // Gets the user data from the presentation.
        initCompilerParameter();
        // Creates the functional tasks.
        speedoWizardCompiler.init();
        // Runs the functional tasks.
        speedoWizardCompiler.process();
        // TODO
        return true;
    }
}
